/*  Compiler of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <lex.h>
#include <compile.h>



static char *instr_name[] = { "TA", "TC", "DR", "ST", "JZ", "JMP", "BOP", "UNOP", "WRT",
	"BOPS", "STS", "TCS", "FD", "BK", "LFT", "RGT", "PU", "PD", "HLT" };


static struct Instruction mem[MAX_PROG_LEN];
static int PC = 0; /* program counter */


extern LexAnalyzer *lex;

/* generates a fragment of program, useful for expressions */
int
gen_prog_fragment( struct Instruction * frag, size_t len )
{
	size_t i;
	int pc = 0;
	for ( i = 0; i < len; i++ ) {
		pc = gen_opcode( frag[i] . opcode, frag[i] . operand );
	}

	return pc;
}

/* generates opcode with potential operand */
int
gen_opcode( opcode_t opcode, int operand )
{
	if ( PC >= MAX_PROG_LEN ) {
		fprintf( stderr, __FILE__ ": Sorry, maximum lenght of program has exceeded! (%d)\n", MAX_PROG_LEN);
		exit( EXIT_FAILURE );
	}
	mem[PC].opcode    = opcode;
	mem[PC].operand = operand;

	return PC++;
}


/* Sets actual PC value as an operand of instruction on address 'address'.
   Useful for JUMPS */
void
setPC_as_operand( int address )
{
	if ( address < 0 || address >= MAX_PROG_LEN ) {
		fprintf( stderr, __FILE__ ": Sorry, maximum lenght of program has exceeded! (%d)\n", MAX_PROG_LEN);
		exit( EXIT_FAILURE );
	}

	mem[address].operand = PC;
	switch ( mem[address].opcode ) {
		case JZ:
		case JMP:
			break;
		default:
			fprintf(stderr, __FILE__ ": Warning: Are you sure? No JZ or JMP there ... \n");
	}
	
}

int
getPC(void)
{
	return PC;
}

/* memory dissassemble */
void
dissassemble( FILE *fp )
{
	int i;
	if ( !fp ) {
		fprintf( stderr, __FILE__ ": FP == NULL!\n" );
		return ;
	}

	fprintf( fp, "Dissassembling memory ... (PC=0x%x) \n", PC );

	for ( i=0; i<PC; i++ ){
		switch ( mem[i].opcode ) {
			case TA:
			case TC:
			case TCS:
			case JMP:
			case JZ:
			case BOPS:
				/* these has one operand */
				fprintf( fp, "0x%x\t%s\t%d\n", i, instr_name[mem[i].opcode], mem[i].operand );
				break;
			case BOP:
				fprintf( fp, "0x%x\t%s\t%s\n", i, instr_name[mem[i].opcode], lex -> getTokenName( (lexsym_t) mem[i].operand ) );
				break;
			case UNOP:
				fprintf( fp, "0x%x\t%s\t%s\n", i, instr_name[mem[i].opcode], lex -> getTokenName( (lexsym_t) mem[i].operand ) );
				break;
			case DR:
			case ST:
			case WRT:
			case STS:
			case FD:
			case BK:
			case LFT:
			case RGT:
			case PU:
			case PD:
			case HLT:
				/* these has only one operand */
				fprintf( fp, "0x%x\t%s\n", i, instr_name[mem[i].opcode]);
				break;
			default:
				fprintf( stderr, "Unknown OPCODE! (%d)\n", mem[i].opcode);
				exit( EXIT_FAILURE );

		}
	}
}

void
dissassemble( struct Instruction *mem, int len, FILE *fp )
{
	int i;
	if ( !fp ) {
		fprintf( stderr, __FILE__ ": FP == NULL!\n" );
		return ;
	}

	fprintf( fp, "Dissassembling memory ... (PC=0x%x) \n", len );

	for ( i=0; i<len; i++ ){
		switch ( mem[i].opcode ) {
			case TA:
			case TC:
			case TCS:
			case JMP:
			case JZ:
			case BOPS:
				/* these has one operand */
				fprintf( fp, "0x%x\t%s\t%d\n", i, instr_name[mem[i].opcode], mem[i].operand );
				break;
			case BOP:
				fprintf( fp, "0x%x\t%s\t%s\n", i, instr_name[mem[i].opcode], lex -> getTokenName( (lexsym_t) mem[i].operand ) );
				break;
			case UNOP:
				fprintf( fp, "0x%x\t%s\t%s\n", i, instr_name[mem[i].opcode], lex -> getTokenName( (lexsym_t) mem[i].operand ) );
				break;
			case DR:
			case ST:
			case WRT:
			case STS:
			case FD:
			case BK:
			case LFT:
			case RGT:
			case PU:
			case PD:
			case HLT:
				/* these has only one operand */
				fprintf( fp, "0x%x\t%s\n", i, instr_name[mem[i].opcode]);
				break;
			default:
				fprintf( stderr, "Unknown OPCODE! (%d)\n", mem[i].opcode);
				exit( EXIT_FAILURE );

		}
	}
}


void
save_memory( FILE *fp )
{

	u_int32_t buf;

	int i;
	for ( i=0; i<PC; i++ ) {
		switch ( mem[i].opcode ) {
			case TA:
			case TC:
			case TCS:
			case BOP:
			case UNOP:
			case JMP:
			case JZ:
			case BOPS:
			case DR:
			case ST:
			case WRT:
			case STS:
			case FD:
			case BK:
			case LFT:
			case RGT:
			case PU:
			case PD:
			case HLT:
				/* all known  opcodes */
				/* This is because we don;t know how opcode_t is interpreted ... So 4B are more explicit,
				   first is opcode, then its operand */
				buf = (u_int32_t) mem[i]. opcode;
				fwrite( & buf, sizeof(u_int32_t), 1, fp );
				buf = (u_int32_t) mem[i]. operand;
				fwrite( & buf, sizeof(u_int32_t), 1, fp );
				break;
			default:
				fprintf( stderr, "Unknown OPCODE! (%d)\n", mem[i].opcode);
				exit( EXIT_FAILURE );
		}
	}
		
}
