#ifndef _COMPILE_H_
#define _COMPILE_H_


/* Instructions:
   TA addr  - transfer address
   TC const - transfer const
   DR       - derefernce
   ST       - store
   JZ addr  - jump if zero
   JMP addr - jump now
   BOP      - binary operation
   UNOP     - unary operation
   WRT      - write
   BOPS     - binary operation on strings
   STS      - store string
   TCS      - transfer string const
   FD       - forward
   BK       - back
   LT       - left
   RT       - right
   PU       - penup
   PD       - pendown
 */
typedef enum {
	TA, TC, DR, ST, JZ, JMP, BOP, UNOP, WRT,
	BOPS, STS, TCS, 
	FD, BK, LFT, RGT, PU, PD, HLT
} opcode_t;



#define MAX_PROG_LEN 1024  /* instructions */

struct Instruction {
	opcode_t opcode;
	int operand;
};

int
gen_prog_fragment( struct Instruction *, size_t len );

int
gen_opcode( opcode_t, int operand = 0 );

void setPC_as_operand( int addr );
int getPC(void);

void dissassemble( FILE *fp = stdout );
void dissassemble( struct Instruction *mem, int len, FILE *fp );
void save_memory( FILE *fp );

#endif
