/*  Expressions of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <lex.h>
#include <expr.h>
#include <var.h>
#include <compile.h>

#define BINOP BOP

extern LexAnalyzer *lex;
extern VarStorage  *var;

Expr::Expr( )
{
#ifdef EXPRDEBUG
	printrules = 1;
#else
	printrules = 0;
#endif

	len = 0;
	compiled_expr = new Instruction[MAX_COMPILED_EXPR_LEN]; 
}

Expr::~Expr()
{
}

/* Gets compiled exression */
struct Instruction *
Expr::getCompiledExpr()
{
	return compiled_expr;
}

size_t 
Expr::getCompiledExprLen()
{
	return len;
}

/* Adds an instruction to the compiled expression fragment */
void
Expr::addInstr( opcode_t opcode, int operand )
{
	compiled_expr[len].opcode  = opcode;
	compiled_expr[len].operand = operand;
	len++;
}

void
Expr::Start( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		case BIT_NEG:
		case MINUS:
		case LOGIC_NOT:
		/* (1) S -> A=S	 [n(~-!] */
			if( printrules )
			printf(" (1) S -> A=S	 [n(~-!] \n");
			Logic();
			if ( lex -> getActualToken() -> lextype == SEMICOLON ) {
				/* Ok, parsed. The expression must end with ';' here
				   or continue with '=' */
				break;
			}
			lex -> compareLexToken( ASSIGN );
			Start();
			addInstr( BINOP, ASSIGN );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error! Only IDENT, NUMB, DOUBLE_QUOTE, ( ~ - ! are permited\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}


/* Logic Operators category : && || @ == <= >= < > ! */
void 
Expr::Logic( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		case BIT_NEG:
		case MINUS:
		/* (2) A -> PC	 [n(~-] */
			if( printrules )
			printf(" (2) A -> PC	 [n(~-] \n");
			Plus();
			LogicRest();
			break;
		case LOGIC_NOT:
		/* (3) A -> !PC	 [!] */
			if( printrules )
			printf(" (3) A -> !PC	 [!] \n");
			lex -> compareLexToken( LOGIC_NOT );
			Plus();
			/*** ASM ***/
			addInstr( UNOP, LOGIC_NOT );
			LogicRest();
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error! Only IDENT NUMB DOUBLE_QUOTE ( ~ - ! are permited here.\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Expr::LogicRest( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case LOGIC_AND:
		/* (4) C -> aPC	 [a] */
			if( printrules )
			printf(" (4) C -> &&PC	 [&&] \n");
			lex -> compareLexToken( LOGIC_AND );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, LOGIC_AND );
			
			LogicRest();
			break;
		case LOGIC_OR:
		/* (5) C -> oPC	 [o] */
			if( printrules )
			printf(" (5) C -> ||PC	 [||] \n");
			lex -> compareLexToken( LOGIC_OR );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, LOGIC_OR );
			
			LogicRest();
			break;
		case LOGIC_XOR:
		/* (6) C -> xPC	 [x] */
			if( printrules )
			printf(" (6) C -> @PC	 [@] \n");
			lex -> compareLexToken( LOGIC_XOR );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, LOGIC_XOR );
			
			LogicRest();
			break;
		case EQ:
		/* (7) C -> qPC	 [q] */
			if( printrules )
			printf(" (7) C -> ==PC	 [==] \n");
			lex -> compareLexToken( EQ );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, EQ );
			
			LogicRest();
			break;
		case GE:
		/* (8) C -> hPC	 [h] */
			if( printrules )
			printf(" (8) C -> >=PC	 [>=] \n");
			lex -> compareLexToken( GE );
			Plus();

			/*** ASM ***/
			addInstr( BINOP, GE );

			LogicRest();
			break;
		case LE:
		/* (9) C -> mPC	 [m] */
			if( printrules )
			printf(" (9) C -> <=PC	 [<=] \n");
			lex -> compareLexToken( LE );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, LE );

			LogicRest();
			break;
		case GT:
		/* (10) C -> gPC	 [g] */
			if( printrules )
			printf(" (10) C -> gPC	 [g] \n");
			lex -> compareLexToken( GT );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, GT );
			
			LogicRest();
			break;
		case LT:
		/* (11) C -> tPC	 [t] */
			if( printrules )
			printf(" (11) C -> tPC	 [t] \n");
			lex -> compareLexToken( LT );
			Plus();
			
			/*** ASM ***/
			addInstr( BINOP, LT );
			
			LogicRest();
			break;
		case EOI:
		case SEMICOLON:
		case ASSIGN:
		case RPAR:
		/* (12) C -> e	 [e=)] */
			if( printrules )
			printf(" (12) C -> e	 [e=)] \n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::LogicRest(): Input parse error! Only && || @ == >= <= < > EOI ; = ) are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

/* Plus Operators category: + - */
void
Expr::Plus( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		case BIT_NEG:
		/* (13) P -> BQ	 [n(~] */
			if( printrules )
			printf(" (13) P -> BQ	 [n(~] \n");
			Bit();
			PlusRest();
			break;
		case MINUS:
		/* (14) P -> -BQ	 [-] */
			if( printrules )
			printf(" (14) P -> -BQ	 [-] \n");
			lex -> compareLexToken( MINUS );
			Bit();
			
			/*** ASM ***/
			addInstr( UNOP, MINUS );
			
			PlusRest();
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::Plus(): Input parse error! Only IDENT NUMB DOUBLE_QUOTE ( ~ - are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Expr::PlusRest( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case PLUS:
		/* (15) Q -> +BQ	 [+] */
			if( printrules )
			printf(" (15) Q -> +BQ	 [+] \n");
			lex -> compareLexToken( PLUS );
			Bit();
			
			/*** ASM ***/
			addInstr( BINOP, PLUS );
			
			PlusRest();
			break;
		case MINUS:
		/* (16) Q -> -BQ	 [-] */
			if( printrules )
			printf(" (16) Q -> -BQ	 [-] \n");
			lex -> compareLexToken( MINUS );
			Bit();
			
			/*** ASM ***/
			addInstr( BINOP, MINUS );
			
			PlusRest();
			break;
		case EOI:
		case SEMICOLON:
		case LOGIC_AND:
		case LOGIC_OR:
		case LOGIC_XOR:
		case EQ:
		case GE:
		case LE:
		case GT:
		case LT:
		case ASSIGN:
		case RPAR:
		/* (17) Q -> e	 [eaoxqhmgt=)] */
			if( printrules )
			printf(" (17) Q -> e	 [eaoxqhmgt=)] \n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::PlusRest(): Input parse error! Only + - EOI ; && || @ == >= <= < > = ) are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}


/* Bit Operators category: ~ & | ^ << >> */
void 
Expr::Bit( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		/* (18) B -> MD	 [n(] */
			if( printrules )
			printf(" (18) B -> MD	 [n(] \n");
			Multiply();
			BitRest();
			break;
		case BIT_NEG:
		/* (19) B -> ~MD	 [~] */
			if( printrules )
			printf(" (19) B -> ~MD	 [~] \n");
			lex -> compareLexToken( BIT_NEG );
			Multiply();
			
			/*** ASM ***/
			addInstr( UNOP, BIT_NEG );

			BitRest();
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::Bit(): Input parse error! Only DOUBLE_QUOTE IDENT NUMB ( ~ are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Expr::BitRest( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case LEFT_SHIFT:
		/* (20) D -> lMD	 [l] */
			if( printrules )
			printf(" (20) D -> <<MD	 [<<] \n");
			lex -> compareLexToken( LEFT_SHIFT );
			Multiply();
			
			/*** ASM ***/
			addInstr( BINOP, LEFT_SHIFT );
			
			BitRest();
			break;
		case RIGHT_SHIFT:
		/* (21) D -> rMD	 [r] */
			if( printrules )
			printf(" (21) D -> >>MD	 [>>] \n");
			lex -> compareLexToken( RIGHT_SHIFT );
			Multiply();
			
			/*** ASM ***/
			addInstr( BINOP, RIGHT_SHIFT );
			
			BitRest();
			break;
		case BIT_OR:
		/* (22) D -> |MD	 [|] */
			if( printrules )
			printf(" (22) D -> |MD	 [|] \n");
			lex -> compareLexToken( BIT_OR );
			Multiply();
			
			/*** ASM ***/
			addInstr( BINOP, BIT_OR );
			
			BitRest();
			break;
		case BIT_AND:
		/* (23) D -> &MD	 [&] */
			if( printrules )
			printf(" (23) D -> &MD	 [&] \n");
			lex -> compareLexToken( BIT_AND );
			Multiply();
			
			/*** ASM ***/
			addInstr( BINOP, BIT_AND );
			
			BitRest();
			break;
		case BIT_XOR:
		/* (24) D -> ^MD	 [^] */
			if( printrules )
			printf(" (24) D -> ^MD	 [^] \n");
			lex -> compareLexToken( BIT_XOR );
			Multiply();
			
			/*** ASM ***/
			addInstr( BINOP, BIT_XOR );
			
			BitRest();
			break;
		case EOI:
		case SEMICOLON:
		case PLUS:
		case MINUS:
		case LOGIC_AND:
		case LOGIC_OR:
		case LOGIC_XOR:
		case EQ:
		case GE:
		case LE:
		case GT:
		case LT:
		case ASSIGN:
		case RPAR:
		/* (25) D -> e	 [e+-aoxqhmgt=)] */
			if( printrules )
			printf(" (25) D -> e	 [e+-aoxqhmgt=)] \n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::BitRest(): Input parse error! Only << >> & | ^ EOI ; + - && || @ == >= <= > < = ) are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}


/* Multiply Operators category : * / % */
void 
Expr::Multiply( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		/* (26) M -> FN	 [n(] */
			if( printrules )
			printf(" (26) M -> FN	 [n(] \n");
			Atom();
			MultiplyRest();
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::Multiply(): Input parse error! Only DOUBLE_QUOTE IDENT NUMB ( are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Expr::MultiplyRest( void )
{
	switch ( lex -> getActualToken() -> lextype ) {
		case MULTIPLY:
		/* (27) N -> *FN	 [*] */
			if( printrules )
			printf(" (27) N -> *FN	 [*] \n");
			lex -> compareLexToken( MULTIPLY );
			Atom();
			
			/*** ASM ***/
			addInstr( BINOP, MULTIPLY );
			
			MultiplyRest();
			break;
		case DIVIDE:
		/* (28) N -> dFN	 [d] */
			if( printrules )
			printf(" (28) N -> /FN	 [/] \n");
			lex -> compareLexToken( DIVIDE );
			Atom();
			
			/*** ASM ***/
			addInstr( BINOP, DIVIDE );
			
			MultiplyRest();
			break;
		case MODULO:
		/* (29) N -> %FN	 [%] */
			if( printrules )
			printf(" (29) N -> %%FN	 [%%] \n");
			lex -> compareLexToken( MODULO );
			Atom();
			
			/*** ASM ***/
			addInstr( BINOP, MODULO );
			
			MultiplyRest();
			break;
		case EOI:
		case SEMICOLON:
		case LEFT_SHIFT:
		case RIGHT_SHIFT:
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		case BIT_AND:
		case BIT_XOR:
		case PLUS:
		case MINUS:
		case LOGIC_AND:
		case LOGIC_OR:
		case LOGIC_XOR:
		case EQ:
		case GE:
		case LE:
		case GT:
		case LT:
		case ASSIGN:
		case RPAR:
		/* (30) N -> e	 [elrn(&^+-aoxqhmgt=)] */
			if( printrules )
			printf(" (30) N -> e	 [elrn(&^+-aoxqhmgt=)] \n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::MultiplyRest(): Input parse error! Only * / %% EOI ; << >> IDENT NUMB ( & | ^ + - && || @ >= <= = )are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}


/* Basic Atom : IDENT(Identifier of a variable) 
   or NUMB (Number constant) or DOUBLE_QUOTE (string constant)*/
void 
Expr::Atom( void )
{
	struct Token *ot = lex -> getActualToken();
	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
			/* (31) F -> n	 [n] */
			if( printrules )
			printf(" (31) F -> n	 [n] \n");
			lex -> compareAtom();
			addInstr( TA, var -> getAddr( ot -> ident ) ) ;
			break;

		case NUMB:
		/* (31) F -> n	 [n] */
			if( printrules )
			printf(" (31) F -> n	 [n] \n");
			lex -> compareAtom();
			addInstr( TC, ot -> numb ) ;
			break;

		case DOUBLE_QUOTE:
		/* (31) F -> n	 [n] */
			if( printrules )
			printf(" (31) F -> n	 [n] \n");
			lex -> compareAtom();
			break;
		case LPAR:
		/* (32) F -> (A)	 [(] */
			if( printrules )
			printf(" (32) F -> (A)	 [(] \n");
			lex -> compareLexToken( LPAR );
			Logic();
			lex -> compareLexToken( RPAR );
			break;
		default:
			fprintf(stderr, __FILE__ ": Expr::Atom(): Input parse error! Only DOUBLE_QUOTE IDENT NUMB ( are permited here\n");
			fprintf(stderr,"Got %d\n", lex -> getActualToken() -> lextype );
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

