/* Using this as a renderer for MIKC language 
 * Copyright (C) 2003 Michal Conos 
 */

/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include "gtk-draw.h"

/* Backing pixmap for drawing area */
static GdkPixmap *pixmap = NULL;
static GtkWidget *drawing_area;

static int WIN_WIDTH;
static int WIN_HEIGHT;


/* Create a new backing pixmap of the appropriate size */
static gint configure_event( GtkWidget         *widget,
                             GdkEventConfigure *event )
{
  if (pixmap)
    gdk_pixmap_unref(pixmap);

  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height,
			  -1);
  gdk_draw_rectangle (pixmap,
		      widget->style->white_gc,
		      TRUE,
		      0, 0,
		      widget->allocation.width,
		      widget->allocation.height);

  return TRUE;
}

/* Redraw the screen from the backing pixmap */
static gint expose_event( GtkWidget      *widget,
                          GdkEventExpose *event )
{
  gdk_draw_pixmap(widget->window,
		  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		  pixmap,
		  event->area.x, event->area.y,
		  event->area.x, event->area.y,
		  event->area.width, event->area.height);

  return FALSE;
}

/* Draw a rectangle on the screen */
void draw_rectangle( gint x, gint y, gint w, gint h)
{
  GdkRectangle update_rect;

  update_rect.x = x;
  update_rect.y = y;
  update_rect.width = w;
  update_rect.height = h;
  gdk_draw_rectangle (pixmap,
		      drawing_area->style->black_gc,
		      FALSE,
		      update_rect.x, update_rect.y,
		      update_rect.width, update_rect.height);
  gtk_widget_draw (drawing_area, &update_rect);
}

/* Draw a rectangle on the screen */
void draw_filled_rectangle( gint x, gint y, gint w, gint h)
{
  GdkRectangle update_rect;

  update_rect.x = x;
  update_rect.y = y;
  update_rect.width = w;
  update_rect.height = h;
  gdk_draw_rectangle (pixmap,
		      drawing_area->style->black_gc,
		      TRUE,
		      update_rect.x, update_rect.y,
		      update_rect.width, update_rect.height);
  gtk_widget_draw (drawing_area, &update_rect);
}

/* clear window */
void clear_window( void )
{
  GdkRectangle update_rect;

  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = WIN_WIDTH;
  update_rect.height = WIN_HEIGHT;
  gdk_draw_rectangle (pixmap,
		      drawing_area->style->white_gc,
		      TRUE,
		      update_rect.x, update_rect.y,
		      update_rect.width, update_rect.height);
  gtk_widget_draw (drawing_area, &update_rect);
}


/* draw line */
void draw_line( gint x1, gint y1, gint x2, gint y2)
{
  GdkRectangle update_rect;
  gint sx, sy, w, h;

  w = abs(x1 - x2);
  h = abs(y1 - y2);
  
  if ( x1 < x2 ) sx = x1; else sx = x2;
  if ( y1 < y2 ) sy = y1; else sy = y2;
  
  update_rect.x = sx;
  update_rect.y = sy;
  update_rect.width = w;
  update_rect.height = h;
  gdk_draw_line (pixmap,
		      drawing_area->style->black_gc,
		      x1, y1, x2, y2 );
  gtk_widget_draw (drawing_area, &update_rect);
}

void draw_pixel( gint x, gint y )
{
  GdkRectangle update_rect;
  update_rect.x = x;
  update_rect.y = y;
  update_rect.width = 1;
  update_rect.height = 1;
	
  gdk_draw_point (pixmap,
		      drawing_area->style->black_gc,
		      x, y );
  gtk_widget_draw (drawing_area, &update_rect);
}

void quit ()
{
  gtk_exit (0);
}

  


void
open_window( int   argc, char *argv[], int width, int height )
{
  GtkWidget *window;
  GtkWidget *vbox;

  GtkWidget *button;

  gtk_init (&argc, &argv);

  WIN_WIDTH  = width;
  WIN_HEIGHT = height;

  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_name (window, "Test Input");

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (window), vbox);
  gtk_widget_show (vbox);

  gtk_signal_connect (GTK_OBJECT (window), "destroy",
		      GTK_SIGNAL_FUNC (quit), NULL);

  /* Create the drawing area */

  drawing_area = gtk_drawing_area_new ();
  gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), width, height);
  gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

  gtk_widget_show (drawing_area);

  /* Signals used to handle backing pixmap */

  gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
		      (GtkSignalFunc) expose_event, NULL);
  gtk_signal_connect (GTK_OBJECT(drawing_area),"configure_event",
		      (GtkSignalFunc) configure_event, NULL);

  /* Event signals */

  /*
  gtk_signal_connect (GTK_OBJECT (drawing_area), "motion_notify_event",
		      (GtkSignalFunc) motion_notify_event, NULL);
  gtk_signal_connect (GTK_OBJECT (drawing_area), "button_press_event",
		      (GtkSignalFunc) button_press_event, NULL);
			  */

  gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
			 | GDK_LEAVE_NOTIFY_MASK
			 | GDK_BUTTON_PRESS_MASK
			 | GDK_POINTER_MOTION_MASK
			 | GDK_POINTER_MOTION_HINT_MASK);


  /* .. And a quit button */
  button = gtk_button_new_with_label ("Quit");
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 0);

  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (window));
  gtk_widget_show (button);

  gtk_widget_show (window);
}


  

void
call_gtk_main(void)
{
  gtk_main ();
}
/* example-end */
