/*  Lexical Analyzer of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <lex.h>


static char *TokenNames[] = {
	"IDENT", "NUMB", "PLUS", "MINUS", "MULTIPLY", "DIVIDE", "MODULO", "ASSIGN", "LPAR", "RPAR",
	"LEFT_SHIFT", "RIGHT_SHIFT", "BIT_AND", "BIT_OR", "BIT_NEG", "BIT_XOR", "EQ", "GT", "LT", "GE", "LE",
	"LOGIC_AND", "LOGIC_OR", "LOGIC_XOR", "LOGIC_NOT", "COMMA", "COLON", "SEMICOLON", "DOUBLE_QUOTE",
	"IF", "ELSE", "FI", "WHILE", "DO", "DONE", "FOR", "REPEAT", "UNTIL", "FORWARD", "BACKWARD", "LEFT", "RIGHT",
	"PENUP", "PENDOWN", "PRINT", "EMPTY", "EOI"
};




/* Open desired file or quit  */
LexAnalyzer::LexAnalyzer( char *filename )
{

	
	if ( filename ) {
		fp = fopen( filename, "r" );
		if ( ! fp ) {
			perror("fopen()");
			exit( EXIT_FAILURE );
		}

		fseek( fp, 0L, SEEK_SET );
	}

	bufpos = 0;
	tokenize_was_called = 0;
	tokpos = 0;
}

/* close opened file */
LexAnalyzer::~LexAnalyzer()
{
	fclose( fp );
}


/* print some info about an error which has occured */
void
LexAnalyzer::TokenError()
{
	/* Not sure whether this is right use of snprintf(3) ... */
	char str[100];
	snprintf( (char*)str, sizeof(str), "Line %d: ", actual_token -> line + 1 );
	fprintf( stderr, "%s", str );
	print_details( get_nth_line( actual_token -> line ), actual_token -> column + strlen(str) );
}

/* Returns Token name */
char *
LexAnalyzer::getTokenName( lexsym_t token )
{
	return TokenNames[token];
}

/* compare on lexical element */
void
LexAnalyzer::compareLexToken( lexsym_t symbol )
{
	if( actual_token->lextype == symbol ) {
		(void)getNextToken();
	} else {
		fprintf( stderr, __FILE__ ": Uh-och, syntax error! Can't analyze input!\n");
		fprintf( stderr, "Actual token id:  %d (%s) , got %d (%s) to compare!\n", actual_token->lextype, TokenNames[actual_token->lextype], symbol, TokenNames[symbol] );
		fprintf( stderr, "Token position:%d Line:%d Col:%d\n", tokpos, actual_token->line, actual_token->column );
		TokenError();
		exit( EXIT_FAILURE );
	}
			
}

/* Compares on Atom */
void
LexAnalyzer::compareAtom( )
{
	if ( actual_token->lextype == IDENT || actual_token -> lextype == NUMB || actual_token -> lextype == DOUBLE_QUOTE ) {
		(void) getNextToken();
	} else {
		fprintf( stderr, __FILE__ ": Uh-och, syntax error! Can't analyze input!\n");
		exit( EXIT_FAILURE );
	}
}



/* Lexical analyzer, parse input */
int
LexAnalyzer::tokenize()
{
	size_t N;
	char *line;
	lexsym_t tokid;
	size_t i,len;
	int line_no = 0;
	char *ident, *numb;
	
	bufpos = 0;
	tokenize_was_called = 1;
	
	if ( !fp ) return 0;

	while( N=0, line=NULL, getline( &line, &N, fp ) != -1 ) {
		i = 0;
		printf("line=|%s|\n", line );
		while  ( i < strlen(line) ) {
			if ( ( len=compare( line+i, "forward" ) ) ) {
				tokid = FORWARD;
				printf("!!!FORWARD!!!\n");
			}
			else
			if ( ( len=compare( line+i, "until" ) ) )
				tokid = UNTIL;
			else
			if ( ( len=compare( line+i, "print" ) ) )
				tokid = PRINT;
			else
			if ( ( len=compare( line+i, "if" ) ) )
				tokid = IF;
			else
			if ( ( len=compare( line+i, "else" ) ) )
				tokid = ELSE;
			else
			if ( ( len=compare( line+i, "fi" ) ) )
				tokid = FI;
			else
			if ( ( len=compare( line+i, "while" ) ) )
				tokid = WHILE;
			else
			if ( ( len=compare( line+i, "done" ) ) )
				tokid = DONE;
			else
			if ( ( len=compare( line+i, "for" ) ) )
				tokid = FOR;
			else
			if ( ( len=compare( line+i, "repeat" ) ) )
				tokid = REPEAT;
			else
			if ( ( len=compare( line+i, "do" ) ) )
				tokid = DO;
			else
			if ( ( len=compare( line+i, "back" ) ) )
				tokid = BACKWARD;
			else
			if ( ( len=compare( line+i, "left" ) ) )
				tokid = LEFT;
			else
			if ( ( len=compare( line+i, "right" ) ) )
				tokid = RIGHT;
			else
			if ( ( len=compare( line+i, "penup" ) ) )
				tokid = PENUP;
			else
			if ( ( len=compare( line+i, "pendown" ) ) )
				tokid = PENDOWN;
			else
			if ( ( len=compare( line+i, "==" ) ) )
				tokid = EQ;
			else
			if ( ( len=compare( line+i, ">=" ) ) )
				tokid = GE;
			else
			if ( ( len=compare( line+i, "<=" ) ) )
				tokid = LE;
			else
			if ( ( len=compare( line+i, "&&" ) ) )
				tokid = LOGIC_AND;
			else
			if ( ( len=compare( line+i, "||" ) ) )
				tokid = LOGIC_OR;
			else
			if ( ( len=compare( line+i, "<<" ) ) )
				tokid = LEFT_SHIFT;
			else
			if ( ( len=compare( line+i, ">>" ) ) )
				tokid = RIGHT_SHIFT;
			else
			if ( ( len=compare( line+i, "@" ) ) )
				tokid = LOGIC_XOR;
			else
			if ( ( len=compare( line+i, "!" ) ) )
				tokid = LOGIC_NOT;
			else
			if ( ( len=compare( line+i, "&" ) ) )
				tokid = BIT_AND;
			else
			if ( ( len=compare( line+i, "|" ) ) )
				tokid = BIT_OR;
			else
			if ( ( len=compare( line+i, "~" ) ) )
				tokid = BIT_NEG;
			else
			if ( ( len=compare( line+i, "^" ) ) )
				tokid = BIT_XOR;
			else
			if ( ( len=compare( line+i, "+" ) ) )
				tokid = PLUS;
			else
			if ( ( len=compare( line+i, "-" ) ) )
				tokid = MINUS;
			else
			if ( ( len=compare( line+i, "*" ) ) )
				tokid = MULTIPLY;
			else
			if ( ( len=compare( line+i, "/" ) ) )
				tokid = DIVIDE;
			else
			if ( ( len=compare( line+i, "%" ) ) )
				tokid = MODULO;
			else
			if ( ( len=compare( line+i, "=" ) ) )
				tokid = ASSIGN;
			else
			if ( ( len=compare( line+i, ">" ) ) )
				tokid = GT;
			else
			if ( ( len=compare( line+i, "<" ) ) )
				tokid = LT;
			else
			if ( ( len=compare( line+i, "(" ) ) )
				tokid = LPAR;
			else
			if ( ( len=compare( line+i, ")" ) ) )
				tokid = RPAR;
			else
			if ( ( len=compare( line+i, "," ) ) )
				tokid = COMMA;
			else
			if ( ( len=compare( line+i, ":" ) ) )
				tokid = COLON;
			else
			if ( ( len=compare( line+i, ";" ) ) )
				tokid = SEMICOLON;
			else
			if ( ( len=compare( line+i, "\"" ) ) ) {
				tokid = DOUBLE_QUOTE;
				if ( ( ident = string_constant( line+i+1 ) ) ) {
					tokenbuffer[ bufpos ] . ident = ident;
					len = 2+strlen(ident);
				} else {
					fprintf(stderr, __FILE__ "Have DOUBLE_QUOTE and no string constant! Confused and misleaded!\n");
					exit( EXIT_FAILURE );
				}
			}
			else
			if ( ( ident = could_be_identifier( line+i ) ) ) {
				tokid = IDENT;
				len = strlen(ident);
				printf("Identifier found:|%s|\n", ident );
				tokenbuffer[ bufpos ].ident = ident;
			}
			else
			if ( ( numb = could_be_number( line+i ) ) ) {
				tokid = NUMB;
				len = strlen(numb);
				printf("Number found:|%s|\n", numb );
				tokenbuffer[ bufpos ].numb = strtod( numb, NULL );
			} else 
			if ( line[ i ] == '#' 	) /*skip this line - comment */
				break;
			else
			if ( isspace( line[i] ) ) {
				tokid = EMPTY;
				len = 1;
			} else {
				fprintf(stderr, __FILE__ ": Error: Unknown token on line %d, column %d!\n", line_no, i );
				print_details(line, i);
				exit( EXIT_FAILURE );
			}
			i+=len;
			if ( bufpos >= MAX_TOKENS-1 ) {
				fprintf(stderr, __FILE__ ": Warning: Maximal token storage was exhausted, aborting parsing!\n");
				return 0;
			}
			tokenbuffer[bufpos].lextype = tokid;
			tokenbuffer[bufpos].line    = line_no;
			tokenbuffer[bufpos++].column= i-len;
		}
		line_no++;
	}

	tokenbuffer[bufpos++].lextype = EOI;
	return 1;
}


/* Lexical analyzer, return specified token */
struct Token *
LexAnalyzer::getNextToken()
{	
	if ( !tokenize_was_called ) {
		tokenize();
	}
	
once_again:	
	if ( tokpos >= bufpos ) {
		return NULL;
	}

	if ( tokenbuffer[tokpos] . lextype == EMPTY ) {
		/* skip delimiters */
		tokpos++;
		goto once_again;
	}
		

	actual_token = & tokenbuffer[tokpos];

	return &tokenbuffer[tokpos++];
}

/* Gets actual token */
struct Token *
LexAnalyzer::getActualToken()
{
	printf("***tokid:%d (%s)\n", actual_token->lextype, getTokenName(actual_token->lextype));
	return actual_token;
}

void
LexAnalyzer::resetLexan()
{
	tokpos = 0;
}

/* if s2 is a substring in s1 from the beginning of s1, then returns 1,
   otherwise 0 */
int 
LexAnalyzer::compare( char *s1, char *s2 )
{
	int len = 0;
	if ( !s1 || !s2 )                 return 0;
	if ( !strlen(s1) || !strlen(s2) ) return 0;
	if ( strlen(s2) > strlen(s1) )    return 0;

	while ( *s1 && *s2 ) {
		if ( *s1 != *s2 ) return 0;
		s1++; s2++;
		len ++;
	}

	return len;
}

/* When the specified string contains a legal identifier name
   from its begining, this function will return it, otherwise
   it returns NULL */
char *
LexAnalyzer::could_be_identifier( char *s )
{
	char *pos = s;
	char *ident;
	int  len  = 0;

	if ( !s || !strlen(s) ) return NULL;

	if ( isalpha( *s ) || *s=='_' ) {
		/* Ok, it is an identfieri, begins with lower/uppercase or '_' */
		while ( isalnum(*s) || *s=='_' )  { len++; s++; }
		if ( !len ) return NULL;
		if ( ( ident = (char*)malloc( len+1 ) ) == NULL ) {
			perror(__FILE__ ": malloc()");
			exit( EXIT_FAILURE );
		}

		strncpy( ident, pos, len );
		ident[len]='\0';

		return ident;
		
	} else
		return NULL;
}

/* When the identspecified string contains a legal number
   from its begining, this function will return it, otherwise
   it returns NULL */
char *
LexAnalyzer::could_be_number( char *s )
{
	char *pos = s;
	char *ident;
	int len = 0;
	int hasdot = 0;
	
	if ( !s || !strlen(s) ) return NULL;

	if ( isdigit( *s ) ) {
		/* Ok, number detected */
		while ( isdigit( *s ) || *s=='.' ) {
			if ( *s=='.' ) {
				if ( !hasdot ) 
					hasdot = 1;
				else
					return NULL;
			} 
			s++;
			len++;
		}

		if ( !len ) return NULL;
		
		if ( ( ident = (char*)malloc( len+1 ) ) == NULL ) {
			perror(__FILE__ " : malloc()");
			exit( EXIT_FAILURE );
		}

		strncpy( ident, pos, len );
		ident[len]='\0';

		return ident;
	
	} else
		return NULL;

}

/* This function returns the string constant, if any */
char *
LexAnalyzer::string_constant( char *s )
{
	int len = 0;
	char *pos = s;
	char *ident;

	if ( !s || !strlen(s) ) return NULL;

	while ( *s ) {
		if ( *s == '"' ) break;
		len++;
		s++;
	}

	if ( !*s ) {
		fprintf(stderr,__FILE__ ": Error: String constant must fit on one line!\n");
		exit( EXIT_FAILURE );
	}	

	if ( !len ) return NULL;

	if ( ( ident = (char*)malloc( len+1 ) ) == NULL ) {
		perror(__FILE__ ": malloc()");
		exit( EXIT_FAILURE );
	}

	strncpy( ident, pos, len );
	ident[len]='\0';

	return ident;
}


/* gets n-th line from the input file */
char *
LexAnalyzer::get_nth_line( int n )
{
	int lineno = 0;
	size_t N;
	char *line;
	
	fseek( fp, 0L, SEEK_SET );
	
	while( N=0, line=NULL, getline( &line, &N, fp ) != -1 ) {
		if ( lineno++ == n ) return line;
	}

	return NULL;
}

void
LexAnalyzer::print_details( char *line, size_t col )
{
	size_t i;
	if ( line == NULL ) return;
	fprintf( stderr, "%s", line );
	for (i=0;i<col;i++) {
		if ( line[i] == '\t' ) 
			fprintf(stderr,"\t");
		else
			fprintf(stderr," ");
	}
	fprintf(stderr,"^- this is probably the place of an error.\n");
}
