#ifndef _LEX_H_
#define _LEX_H_


#include <stdio.h>
#include <stdlib.h>


#define DELIMITERS "\t\n\r "


/* Operator priority :

   =
   !  &&  ||  @  ==  >=  <=  >  <
   +  -
   << >>  &   |  ^   ~
   *  /   %
   (  )
   
 */



typedef enum {
	
	/* identifier, number */ 
	   IDENT,      NUMB,   
	   
	/* +       -       *         /       %       =     (      )  */ 
	   PLUS, MINUS, MULTIPLY, DIVIDE, MODULO, ASSIGN, LPAR, RPAR,   
	/*      <<          >>        &         |       ~        ^  */
	   LEFT_SHIFT, RIGHT_SHIFT, BIT_AND, BIT_OR, BIT_NEG, BIT_XOR,
	/* ==   >   <  >=  <= */
	   EQ, GT, LT, GE, LE,
	/*    &&         ||         @         ! */
	   LOGIC_AND, LOGIC_OR, LOGIC_XOR, LOGIC_NOT,

	/*   ,      :        ;          "       */
	   COMMA, COLON, SEMICOLON, DOUBLE_QUOTE,

	/* if  else  fi  while  do  done  for  repeat */
	   IF, ELSE, FI, WHILE, DO, DONE, FOR, REPEAT, UNTIL,

	/* forward  backward  left  right  penup  pendown */
	   FORWARD, BACKWARD, LEFT, RIGHT, PENUP, PENDOWN, PRINT,

    /* delimiters */
	   EMPTY,

	/* end of input */
	   EOI
	   
} lexsym_t;

#define MAX_TOKENS 256000L



struct Token {
	lexsym_t lextype;
	double numb;
	char * ident;
	int line, column;
};

class LexAnalyzer {
	public:
	  LexAnalyzer( char *filename );
	  ~LexAnalyzer();
	  void TokenError();
	  char *getTokenName(lexsym_t token);
	  struct Token * getNextToken();
	  struct Token * getActualToken();
	  void resetLexan();
	  void compareLexToken( lexsym_t sym );
	  void compareAtom( );
	  int tokenize();
	  

	
	private:
	  FILE *fp;  	  
	  int bufpos;
	  int tokpos;
	  struct Token tokenbuffer[MAX_TOKENS];
	  struct Token *actual_token;
	  int tokenize_was_called;
	  int compare( char *, char * );
	  char *could_be_identifier( char *);
	  char *could_be_number(char *);
	  char *string_constant(char *);
	  char *get_nth_line( int n );
	  void print_details(char *, size_t);
};


#endif
