/*  MIKC language compiler
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 
#include <stdio.h>
#include <stdlib.h>


#include <lex.h>
#include <var.h>
#include <expr.h>
#include <statm.h>
#include <compile.h>

static char *toknames[] = {
	"IDENT", "NUMB", "PLUS", "MINUS", "MULTIPLY", "DIVIDE", "MODULO", "ASSIGN", "LPAR", "RPAR",
	"LEFT_SHIFT", "RIGHT_SHIFT", "BIT_AND", "BIT_OR", "BIT_NEG", "BIT_XOR", "EQ", "GT", "LT", "GE", "LE",
	"LOGIC_AND", "LOGIC_OR", "LOGIC_XOR", "LOGIC_NOT", "COMMA", "COLON", "SEMICOLON", "DOUBLE_QUOTE",
	"IF", "ELSE", "FI", "WHILE", "DO", "DONE", "FOR", "REPEAT", "UNTIL", "FORWARD", "BACKWARD", "LEFT", "RIGHT",
	"PENUP", "PENDOWN", "PRINT", "EMPTY", "EOI"
};


void
printvar( struct VarInfo *var )
{
	printf("%s:", var -> varname );
	if ( var -> vartype == NUMBER ) {
		printf("%.2f\n", var -> numval );
	}
	if ( var -> vartype == STRING ) {
		printf("%s\n", var -> strval );
	}
}


LexAnalyzer *lex;
VarStorage  *var;

int
main( int argc, char *argv[] )
{
	struct Token *token ;
	Statement *statm;

	FILE *output;

	/* Binary output specification */
	output = fopen("bin.out", "wb");
	if  (!output) {
		perror("fopen()");
		exit( EXIT_FAILURE );
	}
	
	if ( argc >= 2 ) {
		/* We've some arguments ... */
		lex = new LexAnalyzer( argv[1] );
	} else {
		lex = new LexAnalyzer( "test.mikc" );
	}
	var = new VarStorage(8);

//	lex -> tokenize(); // This is not neccessary, it is called automatically by getNextToken()
	
	while ( ( token = lex -> getNextToken() ) ) {
		printf("Tok_type:%d (%s) \n", token -> lextype, toknames[ token -> lextype ] );
		if ( token -> lextype == IDENT ) {
			printf("identifier:%s\n", token -> ident );
		}
		if ( token -> lextype == NUMB ) {
			printf("value:%.2f\n", token -> numb );
		}
		if ( token -> lextype == DOUBLE_QUOTE ) {
			printf("string const:%s\n", token -> ident );
		}
	}

	
	lex -> resetLexan();
	
	statm = new Statement();
	statm -> Start();
	
	lex->~LexAnalyzer();

	dissassemble();
	save_memory( output );

	fclose( output );
	return EXIT_SUCCESS;
		
}


