/*  Statements of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

/* Statements are :
 *
 * STATEMENT -> IF STATEMENT | FOR STATEMENT | WHILE STATEMENT | DO STATEMENT | REPEAT STATEMENT | 
 *              EXPRESSION STATEMENT | PRINT STATEMENT | LEFT STATEMENT | ... | BACKWARD STATEMENT | ; | e
 * IF        -> if EXPRESSION STATEMENT ELSE fi
 * ELSE      -> else STATEMENT
 * FOR       -> for EXPRESSION,EXPRESSION,EXPRESSION STATEMENT done
 * WHILE     -> while EXPRESSION STATEMENT done
 * DO        -> do STATEMENT until EXPRESSION
 * REPEAT    -> repeat EXPRESSION STATEMENT done
 * PRINT     -> print EXPRESSION
 * LEFT      -> left EXPRESSION
 * RIGHT     -> right EXPRESSION
 * FORWARD   -> forward EXPRESSION
 * BACKWARD  -> back EXPRESSION
 * EXPRESSION-> "defined elsewhere ;-)"
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <lex.h>
#include <statm.h>
#include <expr.h>
#include <compile.h>


extern LexAnalyzer *lex;


Statement::Statement()
{
	lex -> getNextToken();
}

Statement::~Statement()
{

}


void 
Statement::Start( void )
{
	Expr *e ;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case IF:
		/* (1) S -> IS	 [i] */
			printf(" (1) S -> IS	 [i] \n");
			StatmIF();
			Start();
			break;
		case FOR:
		/* (2) S -> FS	 [f] */
			printf(" (2) S -> FS	 [f] \n");
			StatmFOR();
			Start();
			break;
		case WHILE:
		/* (3) S -> WS	 [w] */
			printf(" (3) S -> WS	 [w] \n");
			StatmWHILE();
			Start();
			break;
		case DO:
		/* (4) S -> DS	 [q] */
			printf(" (4) S -> DS	 [q] \n");
			StatmDO();
			Start();
			break;
		case REPEAT:
		/* (5) S -> RS	 [r] */
			printf(" (5) S -> RS	 [r] \n");
			StatmREPEAT();
			Start();
			break;
		case LPAR:         /* what everything could start an expression */
		case NUMB:
		case IDENT:
		case DOUBLE_QUOTE:
		/* (6) S -> ES	 [c] */
			printf(" (6) S -> ES	 [c] \n");
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			Start();
			break;
		case PRINT:
		/* (7) S -> PS	 [p] */
			printf(" (7) S -> PS	 [p] \n");
			StatmPRINT();
			Start();
			break;
		case LEFT:
		/* (8) S -> VS	 [v] */
			printf(" (8) S -> VS	 [v] \n");
			StatmLEFT();
			Start();
			break;
		case RIGHT:
		/* (9) S -> XS	 [x] */
			printf(" (9) S -> XS	 [x] \n");
			StatmRIGHT();
			Start();
			break;
		case FORWARD:
		/* (10) S -> YS	 [y] */
			printf(" (10) S -> YS	 [y] \n");
			StatmFORWARD();
			Start();
			break;
		case BACKWARD:
		/* (11) S -> ZS	 [z] */
			printf(" (11) S -> ZS	 [z] \n");
			StatmBACKWARD();
			Start();
			break;
		case SEMICOLON:
		/* (12) S -> ;	 [;] */
			printf(" (12) S -> ;	 [;] \n");
			lex -> compareLexToken( SEMICOLON );
			Start();
			break;
		case EOI:
		case ELSE:
		case FI:
		case DONE:
		case UNTIL:
		/* (13) S -> e	 [ejkdu] */
			printf(" (13) S -> e	 [ejkdu] \n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmIF( void )
{
	int JZ_addr;
	Expr *e;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case IF:
		/* (14) I -> iESJk	 [i] */
			printf(" (14) I -> iESJk	 [i] \n");
			lex -> compareLexToken( IF );
			printf("***if\n");
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			JZ_addr = gen_opcode( JZ );
			printf("***then\n");
			Start();
			StatmELSE(JZ_addr);
			lex -> compareLexToken( FI );
			printf("fi\n");
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmELSE( int jump_addr )
{
	int JMP_addr;
	switch ( lex -> getActualToken() -> lextype ) {
		case ELSE:
		/* (15) J -> jS	 [j] */
			printf(" (15) J -> jS	 [j] \n");
			lex -> compareLexToken( ELSE );
			printf("else\n");
			JMP_addr = gen_opcode( JMP );
			setPC_as_operand( jump_addr );
			Start();
			setPC_as_operand( JMP_addr );
			break;
		case EOI:
		case SEMICOLON:
		case FI:
		/* J -> e */
			printf(" J -> e\n" );
			setPC_as_operand( jump_addr );
		//	(void)lex -> getNextToken();
		//	Start();
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmFOR( void )
{
	Expr *init_expr, *cond_expr, *post_expr;
	int JZ_addr, loop_start;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case FOR:
		/* (16) F -> fE,E,ESd	 [f] */
			printf(" (16) F -> fE,E,ESd	 [f] \n");
			lex -> compareLexToken( FOR );
			init_expr = StatmEXPRESSION();
			lex -> compareLexToken( COMMA );
			cond_expr = StatmEXPRESSION();
			lex -> compareLexToken( COMMA );
			post_expr = StatmEXPRESSION();

			
			gen_prog_fragment( init_expr -> getCompiledExpr(), init_expr -> getCompiledExprLen() );
			loop_start = getPC();
			gen_prog_fragment( cond_expr -> getCompiledExpr(), cond_expr -> getCompiledExprLen() );
			JZ_addr = gen_opcode( JZ );

			Start();
			gen_prog_fragment( post_expr -> getCompiledExpr(), post_expr -> getCompiledExprLen() );

			gen_opcode( JMP, loop_start );
			setPC_as_operand( JZ_addr );
			

			
			lex -> compareLexToken( DONE );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmWHILE( void )
{
	Expr * cond_expr;
	int loop_start, JNZ_addr;

	switch ( lex -> getActualToken() -> lextype ) {
		case WHILE:
		/* (17) W -> wESd	 [w] */
			printf(" (17) W -> wESd	 [w] \n");
			lex -> compareLexToken( WHILE );
			
			loop_start = getPC();          // need to know the actual PC position
			
			cond_expr = StatmEXPRESSION();
			gen_prog_fragment( cond_expr -> getCompiledExpr(), cond_expr -> getCompiledExprLen() );
			
			JNZ_addr = gen_opcode( JZ );   // generate condition jump
			
			Start();
			
			gen_opcode( JMP, loop_start );     // this is jump to loop_start
			setPC_as_operand( JNZ_addr );      // here we know the address for the condition jump

			lex -> compareLexToken( DONE );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmDO( void )
{
	Expr *cond_expr;
	int loop_start;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case DO:
		/* (18) D -> qSuE	 [q] */
			printf(" (18) D -> qSuE	 [q] \n");              // Compile scheme :
			lex -> compareLexToken( DO );                   //  loop:
			                                                //        DO {
			loop_start = getPC();                           //         ...
			                                                //         ...
			Start();                                        //       UNTIL RXPR;
			lex -> compareLexToken( UNTIL );                //       JZ PC+2
			cond_expr = StatmEXPRESSION();                  //       JMP loop
			gen_prog_fragment( cond_expr -> getCompiledExpr(), cond_expr -> getCompiledExprLen() );
                                                            /// PC+2:
			gen_opcode( JZ, getPC() + 2 );
			gen_opcode( JMP, loop_start );
			
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmREPEAT( void )
{
	Expr *cond_expr;
	int loop_start, JZ1_addr, JZ2_addr;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case REPEAT:
		/* (19) R -> rESd	 [r] */
			printf(" (19) R -> rESd	 [r] \n");
			lex -> compareLexToken( REPEAT );
			cond_expr = StatmEXPRESSION();
			gen_prog_fragment( cond_expr -> getCompiledExpr(), cond_expr -> getCompiledExprLen() );
			JZ1_addr = gen_opcode( JZ );
			
			loop_start = getPC();
			Start();

		
			gen_prog_fragment( cond_expr -> getCompiledExpr(), cond_expr -> getCompiledExprLen() );
			gen_opcode( TC, 1 );
			gen_opcode( BOP, MINUS );
			JZ2_addr = gen_opcode( JZ );
			gen_opcode( JMP, loop_start );
			setPC_as_operand( JZ1_addr );
			setPC_as_operand( JZ2_addr );
			
			lex -> compareLexToken( DONE );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

Expr *
Statement::StatmEXPRESSION( void )
{
	Expr *e = new Expr();

	switch ( lex -> getActualToken() -> lextype ) {
		case IDENT:
		case NUMB:
		case DOUBLE_QUOTE:
		case LPAR:
		/* (20) E -> c	 [c] */
			printf(" (20) E -> c	 [c] \n");
			e -> Start();
			(void) lex -> getNextToken(); // eat SEMICOLON
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}

	return e;
}

void 
Statement::StatmPRINT( void )
{
	Expr *e;
	
	switch ( lex -> getActualToken() -> lextype ) {
		case PRINT:
		/* (21) P -> pE	 [p] */
			printf(" (21) P -> pE	 [p] \n");
			lex -> compareLexToken( PRINT );
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			gen_opcode( WRT );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmLEFT( void )
{
	Expr *e;
	switch ( lex -> getActualToken() -> lextype ) {
		case LEFT:
		/* (22) V -> vE	 [v] */
			printf(" (22) V -> vE	 [v] \n");
			lex -> compareLexToken( LEFT );
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			gen_opcode( LFT );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmRIGHT( void )
{
	Expr *e;
	switch ( lex -> getActualToken() -> lextype ) {
		case RIGHT:
		/* (23) X -> xE	 [x] */
			printf(" (23) X -> xE	 [x] \n");
			lex -> compareLexToken( RIGHT );
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			gen_opcode( RGT );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmFORWARD( void )
{
	Expr *e;
	switch ( lex -> getActualToken() -> lextype ) {
		case FORWARD:
		/* (24) Y -> yE	 [y] */
			printf(" (24) Y -> yE	 [y] \n");
			lex -> compareLexToken( FORWARD );
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			gen_opcode( FD );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}

void 
Statement::StatmBACKWARD( void )
{
	Expr *e;
	switch ( lex -> getActualToken() -> lextype ) {
		case BACKWARD:
		/* (25) Z -> zE	 [z] */
			printf(" (25) Z -> zE	 [z] \n");
			lex -> compareLexToken( BACKWARD );
			e = StatmEXPRESSION();
			gen_prog_fragment( e -> getCompiledExpr(), e -> getCompiledExprLen() );
			gen_opcode( BK );
			break;
		default:
			fprintf(stderr, __FILE__ ": Input parse error!\n");
			lex -> TokenError();
			exit( EXIT_FAILURE );
	}
}


