/*  Variable storage of MIKC language
 *  Copyright (C) 2003 Michal Conos
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>



#include <lex.h>
#include <var.h>

VarStorage::VarStorage( int space )
{
	varstorage = new VarInfo[space];
	varspace   = space;

	init_service();
		
}

VarStorage::VarStorage( )
{
	varstorage = new VarInfo[DEFAULTSPACE];
	varspace   = DEFAULTSPACE;
	
	init_service();
}

VarStorage::~VarStorage( )
{
	delete [] varstorage;
}


int
VarStorage::getAddr( char *varname )
{
	int i;
	int found = 0;

	if ( !varname ) {
		fprintf( stderr, __FILE__ ": You gave NULL variable name!\n");
		exit( EXIT_FAILURE );
	}
	
	for ( i = 0; i < actpos; i++ ) {
		if ( !strcmp( varname, varstorage[i] . varname ) ) {
			found=1;
			break;
		}
		
	}

	if ( found ) return i;
	else {
		varstorage[actpos] . varname = varname;
	}

	return actpos++;
}


/* Gets variable from the storage */
struct VarInfo *
VarStorage::getVariable( char *varname )
{
	int index,c=0;
	int found = 0;
	index = hash( varname, GET );

		
	while ( !found ) {
		if ( !varstorage[index] . free ) {
			if ( !strcmp( varname, varstorage[index] . varname ) ) {
				found = 1;
				break;
			}
		}
		
		if ( c++ == varspace ) {
			fprintf(stderr, __FILE__ ": Unknown variable name: %s!\n", varname );
			exit( EXIT_FAILURE );
		}
		index++;
		index %= varspace;
	}

	return & varstorage[index];
}

/* Stores a variable into a varstorage array */
void
VarStorage::storeVariable( char *varname, double value )
{
	int index;
	index = hash( varname, STORE );

	printf("varname=%s\n", varname);
	if ( ++stored_num >= varspace ) {
		fprintf( stderr, __FILE__ ": Sorry, maximum variable space exhausted! Exit.\n" );
		exit( EXIT_FAILURE );
	}


	/* somebody is already there */
	while ( !varstorage[index] . free ) {
		if ( !strcmp( varstorage[index] . varname, varname ) ) {
			/* the same name */
			break;
		}

		index++;
		index %= varspace;
	}

	/* store varibale info */
	varstorage[index] . free    = 0;
	varstorage[index] . varname = varname;
	varstorage[index] . vartype = NUMBER;
	varstorage[index] . numval  = value; 
}

/* Stores a variable into a varstorage array */
void
VarStorage::storeVariable( char *varname, char * value )
{
	int index;
	index = hash( varname, STORE );

	if ( ++stored_num >= varspace ) {
		fprintf( stderr, __FILE__ ": Sorry, maximum variable space exhausted! Exit.\n" );
		exit( EXIT_FAILURE );
	}

	/* somebody is already there */
	while ( !varstorage[index] . free ) {
		if ( !strcmp( varstorage[index] . varname, varname ) ) {
			/* the same name */
			break;
		}

		index++;
		index %= varspace;
	}

	/* store varibale info */
	varstorage[index] . free    = 0;
	varstorage[index] . varname = varname;
	varstorage[index] . vartype = STRING;
	varstorage[index] . strval  = value; 
}


/* Hash function for a variable */
int
VarStorage::hash( char *s, opmode_t opmode )
{
	size_t i;
	double w, sum;
	int MAX = 'z';
	int MIN = '0' - 1;
	int index ;

	if ( !s || !strlen(s) ) {
		fprintf( stderr, __FILE__ ": You gave me a NULL pointer instead of variable name or varibale name has zero lenght! Exit.\n");
		exit( EXIT_FAILURE );
	}

	
	/* count hash from varibale name */
	w=sum=0;
	for ( i=0; i<strlen(s); i++ ) {
		
		switch ( s[i] ) {
			case '0' ... '9' :
			case 'a' ... 'z' :
			case 'A' ... 'Z' :
				w += (double)(s[i] - MIN) / ( MAX - MIN )* 0.35 + 0.1;
				sum += w * s[i];
				break;
			case  '_' :
				w += 0.1;
				break;
			default:
				fprintf( stderr, __FILE__ ": This character (%c) is must not be included in a legitimate variable name!\n", s[i] );
				exit( EXIT_FAILURE );
		}
	}

	index = (int) floor( sum );
	index %= varspace;

	if ( opmode == STORE )
	/* if slot is not empty, find empty slot */
	if (! varstorage[index] . free ) {
		fprintf( stderr, __FILE__ ": Colision for %s in hash table with %s, index=%d\n", s, varstorage[index] . varname, index );
		while ( !varstorage[index] . free ) { index ++; index %= varspace; }
	}


	return index;
}



void
VarStorage::init_service()
{
	int i;

	for ( i = 0; i<varspace; i++ ) 
		varstorage[i] . free = 1;

	stored_num = 0;
	actpos = 0;
}
